// Copyright 2014 tsuru authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package iaas

import (
	"github.com/tsuru/config"
	"github.com/tsuru/tsuru/db"
	"github.com/tsuru/tsuru/db/storage"
	"github.com/tsuru/tsuru/errors"
	"github.com/tsuru/tsuru/log"
)

type TemplateData struct {
	Name  string
	Value string
}

type TemplateDataList []TemplateData

func (l TemplateDataList) Len() int           { return len(l) }
func (l TemplateDataList) Swap(i, j int)      { l[i], l[j] = l[j], l[i] }
func (l TemplateDataList) Less(i, j int) bool { return l[i].Name < l[j].Name }

type Template struct {
	Name     string `bson:"_id"`
	IaaSName string
	Data     TemplateDataList
}

func FindTemplate(name string) (*Template, error) {
	coll := template_collection()
	defer coll.Close()
	var template Template
	err := coll.FindId(name).One(&template)
	return &template, err
}

func ExpandTemplate(name string, params map[string]string) (map[string]string, error) {
	template, err := FindTemplate(name)
	if err != nil {
		return nil, err
	}
	templateParams := template.paramsMap()
	delete(params, "template")
	// User params will override template params
	for k, v := range templateParams {
		_, isSet := params[k]
		if !isSet {
			params[k] = v
		}
	}
	return params, nil
}

func ListTemplates() ([]Template, error) {
	coll := template_collection()
	defer coll.Close()
	var templates []Template
	err := coll.Find(nil).Sort("_id").All(&templates)
	return templates, err
}

func DestroyTemplate(name string) error {
	coll := template_collection()
	defer coll.Close()
	return coll.RemoveId(name)
}

func (t *Template) Update(toMerge *Template) error {
	currentMap := t.paramsMap()
	toMergeMap := toMerge.paramsMap()
	delete(toMergeMap, "iaas")
	delete(currentMap, "iaas")
	for k, v := range toMergeMap {
		if v == "" {
			delete(currentMap, k)
		} else {
			currentMap[k] = v
		}
	}
	t.Data = make(TemplateDataList, 0, len(currentMap))
	for k, v := range currentMap {
		t.Data = append(t.Data, TemplateData{Name: k, Value: v})
	}
	return t.Save()
}

func (t *Template) Save() error {
	if t.Name == "" {
		return &errors.ValidationError{Message: "template name cannot be empty"}
	}
	_, err := getIaasProvider(t.IaaSName)
	if err != nil {
		return err
	}
	return t.saveToDB()
}

func (t *Template) saveToDB() error {
	coll := template_collection()
	defer coll.Close()
	_, err := coll.UpsertId(t.Name, t)
	return err
}

func (t *Template) paramsMap() map[string]string {
	params := map[string]string{}
	for _, item := range t.Data {
		params[item.Name] = item.Value
	}
	params["iaas"] = t.IaaSName
	return params
}

func template_collection() *storage.Collection {
	name, err := config.GetString("iaas:collection")
	if err != nil {
		name = "iaas_machines"
	}
	name += "_templates"
	conn, err := db.Conn()
	if err != nil {
		log.Errorf("Failed to connect to the database: %s", err)
	}
	return conn.Collection(name)
}
